package sun.nio.ch;

import java.util.Collections;
import sun.net.ExtendedSocketOptions;
import java.util.HashSet;
import java.nio.channels.NetworkChannel;
import java.net.Inet6Address;
import java.nio.channels.MembershipKey;
import java.nio.channels.AlreadyBoundException;
import java.nio.channels.UnsupportedAddressTypeException;
import java.nio.channels.NotYetConnectedException;
import java.net.PortUnreachableException;
import java.nio.ByteBuffer;
import java.util.Set;
import java.net.Inet4Address;
import java.net.NetworkInterface;
import java.net.StandardSocketOptions;
import java.net.SocketOption;
import java.nio.channels.ClosedChannelException;
import java.io.IOException;
import java.net.StandardProtocolFamily;
import sun.net.ResourceManager;
import java.nio.channels.spi.SelectorProvider;
import java.net.SocketAddress;
import java.net.DatagramSocket;
import java.net.InetSocketAddress;
import java.net.InetAddress;
import java.net.ProtocolFamily;
import java.io.FileDescriptor;
import java.nio.channels.DatagramChannel;

class DatagramChannelImpl extends DatagramChannel implements SelChImpl
{
    private static NativeDispatcher nd;
    private final FileDescriptor fd;
    private final int fdVal;
    private final ProtocolFamily family;
    private volatile long readerThread;
    private volatile long writerThread;
    private InetAddress cachedSenderInetAddress;
    private int cachedSenderPort;
    private final Object readLock;
    private final Object writeLock;
    private final Object stateLock;
    private static final int ST_UNINITIALIZED = -1;
    private static final int ST_UNCONNECTED = 0;
    private static final int ST_CONNECTED = 1;
    private static final int ST_KILLED = 2;
    private int state;
    private InetSocketAddress localAddress;
    private InetSocketAddress remoteAddress;
    private DatagramSocket socket;
    private MembershipRegistry registry;
    private boolean reuseAddressEmulated;
    private boolean isReuseAddress;
    private SocketAddress sender;
    
    public DatagramChannelImpl(SelectorProvider provider) throws IOException {
        super(provider);
        this.readerThread = 0L;
        this.writerThread = 0L;
        this.readLock = new Object();
        this.writeLock = new Object();
        this.stateLock = new Object();
        this.state = -1;
        ResourceManager.beforeUdpCreate();
        try {
            this.family = (Net.isIPv6Available() ? StandardProtocolFamily.INET6 : StandardProtocolFamily.INET);
            this.fd = Net.socket(this.family, false);
            this.fdVal = IOUtil.fdVal(this.fd);
            this.state = 0;
        }
        catch (IOException ex) {
            ResourceManager.afterUdpClose();
            throw ex;
        }
    }
    
    public DatagramChannelImpl(SelectorProvider provider, ProtocolFamily family) throws IOException {
        super(provider);
        this.readerThread = 0L;
        this.writerThread = 0L;
        this.readLock = new Object();
        this.writeLock = new Object();
        this.stateLock = new Object();
        this.state = -1;
        if (family != StandardProtocolFamily.INET && family != StandardProtocolFamily.INET6) {
            if (family == null) {
                throw new NullPointerException("'family' is null");
            }
            throw new UnsupportedOperationException("Protocol family not supported");
        }
        else {
            if (family == StandardProtocolFamily.INET6 && !Net.isIPv6Available()) {
                throw new UnsupportedOperationException("IPv6 not available");
            }
            ResourceManager.beforeUdpCreate();
            try {
                this.family = family;
                this.fd = Net.socket(family, false);
                this.fdVal = IOUtil.fdVal(this.fd);
                this.state = 0;
            }
            catch (IOException ex) {
                ResourceManager.afterUdpClose();
                throw ex;
            }
            return;
        }
    }
    
    public DatagramChannelImpl(SelectorProvider provider, FileDescriptor fd) throws IOException {
        super(provider);
        this.readerThread = 0L;
        this.writerThread = 0L;
        this.readLock = new Object();
        this.writeLock = new Object();
        this.stateLock = new Object();
        this.state = -1;
        ResourceManager.beforeUdpCreate();
        this.family = (Net.isIPv6Available() ? StandardProtocolFamily.INET6 : StandardProtocolFamily.INET);
        this.fd = fd;
        this.fdVal = IOUtil.fdVal(fd);
        this.state = 0;
        this.localAddress = Net.localAddress(fd);
    }
    
    @Override
    public DatagramSocket socket() {
        synchronized (this.stateLock) {
            if (this.socket == null) {
                this.socket = DatagramSocketAdaptor.create(this);
            }
            return this.socket;
        }
    }
    
    @Override
    public SocketAddress getLocalAddress() throws IOException {
        synchronized (this.stateLock) {
            if (!this.isOpen()) {
                throw new ClosedChannelException();
            }
            return Net.getRevealedLocalAddress(this.localAddress);
        }
    }
    
    @Override
    public SocketAddress getRemoteAddress() throws IOException {
        synchronized (this.stateLock) {
            if (!this.isOpen()) {
                throw new ClosedChannelException();
            }
            return this.remoteAddress;
        }
    }
    
    @Override
    public <T> DatagramChannel setOption(SocketOption<T> obj, T t) throws IOException {
        if (obj == null) {
            throw new NullPointerException();
        }
        if (!this.supportedOptions().contains(obj)) {
            throw new UnsupportedOperationException("'" + obj + "' not supported");
        }
        synchronized (this.stateLock) {
            this.ensureOpen();
            if (obj == StandardSocketOptions.IP_TOS || obj == StandardSocketOptions.IP_MULTICAST_TTL || obj == StandardSocketOptions.IP_MULTICAST_LOOP) {
                Net.setSocketOption(this.fd, this.family, obj, t);
                return this;
            }
            if (obj == StandardSocketOptions.IP_MULTICAST_IF) {
                if (t == null) {
                    throw new IllegalArgumentException("Cannot set IP_MULTICAST_IF to 'null'");
                }
                NetworkInterface networkInterface = (NetworkInterface)t;
                if (this.family == StandardProtocolFamily.INET6) {
                    int index = networkInterface.getIndex();
                    if (index == -1) {
                        throw new IOException("Network interface cannot be identified");
                    }
                    Net.setInterface6(this.fd, index);
                }
                else {
                    Inet4Address anyInet4Address = Net.anyInet4Address(networkInterface);
                    if (anyInet4Address == null) {
                        throw new IOException("Network interface not configured for IPv4");
                    }
                    Net.setInterface4(this.fd, Net.inet4AsInt(anyInet4Address));
                }
                return this;
            }
            else {
                if (obj == StandardSocketOptions.SO_REUSEADDR && Net.useExclusiveBind() && this.localAddress != null) {
                    this.reuseAddressEmulated = true;
                    this.isReuseAddress = (boolean)t;
                }
                Net.setSocketOption(this.fd, Net.UNSPEC, obj, t);
                return this;
            }
        }
    }
    
    @Override
    public <T> T getOption(SocketOption<T> obj) throws IOException {
        if (obj == null) {
            throw new NullPointerException();
        }
        if (!this.supportedOptions().contains(obj)) {
            throw new UnsupportedOperationException("'" + obj + "' not supported");
        }
        synchronized (this.stateLock) {
            this.ensureOpen();
            if (obj == StandardSocketOptions.IP_TOS || obj == StandardSocketOptions.IP_MULTICAST_TTL || obj == StandardSocketOptions.IP_MULTICAST_LOOP) {
                return Net.getSocketOption(this.fd, this.family, obj);
            }
            if (obj == StandardSocketOptions.IP_MULTICAST_IF) {
                if (this.family == StandardProtocolFamily.INET) {
                    int interface7 = Net.getInterface4(this.fd);
                    if (interface7 == 0) {
                        return null;
                    }
                    NetworkInterface byInetAddress = NetworkInterface.getByInetAddress(Net.inet4FromInt(interface7));
                    if (byInetAddress == null) {
                        throw new IOException("Unable to map address to interface");
                    }
                    return byInetAddress;
                }
                else {
                    int interface6 = Net.getInterface6(this.fd);
                    if (interface6 == 0) {
                        return null;
                    }
                    NetworkInterface byIndex = NetworkInterface.getByIndex(interface6);
                    if (byIndex == null) {
                        throw new IOException("Unable to map index to interface");
                    }
                    return byIndex;
                }
            }
            else {
                if (obj == StandardSocketOptions.SO_REUSEADDR && this.reuseAddressEmulated) {
                    return Boolean.valueOf(this.isReuseAddress);
                }
                return Net.getSocketOption(this.fd, Net.UNSPEC, obj);
            }
        }
    }
    
    @Override
    public final Set<SocketOption<?>> supportedOptions() {
        return DatagramChannelImpl.DefaultOptionsHolder.defaultOptions;
    }
    
    private void ensureOpen() throws ClosedChannelException {
        if (!this.isOpen()) {
            throw new ClosedChannelException();
        }
        return;
    }
    
    @Override
    public SocketAddress receive(ByteBuffer byteBuffer) throws IOException {
        if (byteBuffer.isReadOnly()) {
            throw new IllegalArgumentException("Read-only buffer");
        }
        if (byteBuffer == null) {
            throw new NullPointerException();
        }
        synchronized (this.readLock) {
            this.ensureOpen();
            if (this.localAddress() == null) {
                this.bind(null);
            }
            int n = 0;
            ByteBuffer temporaryDirectBuffer = null;
            try {
                this.begin();
                if (!this.isOpen()) {
                    __Null _Null3 = null;
                    if (temporaryDirectBuffer != null) {
                        Util.releaseTemporaryDirectBuffer(temporaryDirectBuffer);
                    }
                    this.readerThread = 0L;
                    this.end(n > 0 || n == -2);
                    assert !(!IOStatus.check(n));
                    return _Null3;
                }
                else {
                    SecurityManager securityManager = System.getSecurityManager();
                    this.readerThread = NativeThread.current();
                    if (this.isConnected() || securityManager == null) {
                        do {
                            n = this.receive(this.fd, byteBuffer);
                        } while (n == -3 && this.isOpen());
                        if (n == -2) {
                            __Null _Null2 = null;
                            if (temporaryDirectBuffer != null) {
                                Util.releaseTemporaryDirectBuffer(temporaryDirectBuffer);
                            }
                            this.readerThread = 0L;
                            this.end(n > 0 || n == -2);
                            assert !(!IOStatus.check(n));
                            return _Null2;
                        }
                    }
                    else {
                        temporaryDirectBuffer = Util.getTemporaryDirectBuffer(byteBuffer.remaining());
                        while (true) {
                            n = this.receive(this.fd, temporaryDirectBuffer);
                            if (n == -3 && this.isOpen()) {
                                continue;
                            }
                            if (n == -2) {
                                __Null _Null = null;
                                if (temporaryDirectBuffer != null) {
                                    Util.releaseTemporaryDirectBuffer(temporaryDirectBuffer);
                                }
                                this.readerThread = 0L;
                                this.end(n > 0 || n == -2);
                                assert !(!IOStatus.check(n));
                                return _Null;
                            }
                            else {
                                InetSocketAddress inetSocketAddress = (InetSocketAddress)this.sender;
                                try {
                                    securityManager.checkAccept(inetSocketAddress.getAddress().getHostAddress(), inetSocketAddress.getPort());
                                }
                                catch (SecurityException ex) {
                                    temporaryDirectBuffer.clear();
                                    n = 0;
                                    continue;
                                }
                                temporaryDirectBuffer.flip();
                                byteBuffer.put(temporaryDirectBuffer);
                                break;
                            }
                        }
                    }
                    SocketAddress sender = this.sender;
                    if (temporaryDirectBuffer != null) {
                        Util.releaseTemporaryDirectBuffer(temporaryDirectBuffer);
                    }
                    this.readerThread = 0L;
                    this.end(n > 0 || n == -2);
                    assert !(!IOStatus.check(n));
                    return sender;
                }
            }
            finally {
                if (temporaryDirectBuffer != null) {
                    Util.releaseTemporaryDirectBuffer(temporaryDirectBuffer);
                }
                this.readerThread = 0L;
                this.end(n > 0 || n == -2);
                assert !(!IOStatus.check(n));
            }
        }
    }
    
    private int receive(FileDescriptor fileDescriptor, ByteBuffer byteBuffer) throws IOException {
        int position = byteBuffer.position();
        int limit = byteBuffer.limit();
        assert position <= limit;
        int a = (position <= limit) ? (limit - position) : 0;
        if (byteBuffer instanceof DirectBuffer && a > 0) {
            return this.receiveIntoNativeBuffer(fileDescriptor, byteBuffer, a, position);
        }
        int max = Math.max(a, 1);
        ByteBuffer temporaryDirectBuffer = Util.getTemporaryDirectBuffer(max);
        try {
            int receiveIntoNativeBuffer = this.receiveIntoNativeBuffer(fileDescriptor, temporaryDirectBuffer, max, 0);
            temporaryDirectBuffer.flip();
            if (receiveIntoNativeBuffer > 0 && a > 0) {
                byteBuffer.put(temporaryDirectBuffer);
            }
            return receiveIntoNativeBuffer;
        }
        finally {
            Util.releaseTemporaryDirectBuffer(temporaryDirectBuffer);
        }
    }
    
    private int receiveIntoNativeBuffer(FileDescriptor fileDescriptor, ByteBuffer byteBuffer, int n, int n2) throws IOException {
        int receive0 = this.receive0(fileDescriptor, ((DirectBuffer)byteBuffer).address() + (long)n2, n, this.isConnected());
        if (receive0 > 0) {
            byteBuffer.position(n2 + receive0);
        }
        return receive0;
    }
    
    @Override
    public int send(ByteBuffer byteBuffer, SocketAddress socketAddress) throws IOException {
        if (byteBuffer == null) {
            throw new NullPointerException();
        }
        synchronized (this.writeLock) {
            this.ensureOpen();
            InetSocketAddress checkAddress = Net.checkAddress(socketAddress);
            InetAddress address = checkAddress.getAddress();
            if (address == null) {
                throw new IOException("Target address not resolved");
            }
            synchronized (this.stateLock) {
                if (!this.isConnected()) {
                    if (socketAddress == null) {
                        throw new NullPointerException();
                    }
                    SecurityManager securityManager = System.getSecurityManager();
                    if (securityManager != null) {
                        if (address.isMulticastAddress()) {
                            securityManager.checkMulticast(address);
                        }
                        else {
                            securityManager.checkConnect(address.getHostAddress(), checkAddress.getPort());
                        }
                    }
                }
                else {
                    if (!socketAddress.equals(this.remoteAddress)) {
                        throw new IllegalArgumentException("Connected address not equal to target address");
                    }
                    int write = this.write(byteBuffer);
                    return write;
                }
            }
            int send = 0;
            try {
                this.begin();
                if (!this.isOpen()) {
                    int n = 0;
                    this.writerThread = 0L;
                    this.end(send > 0 || send == -2);
                    assert !(!IOStatus.check(send));
                    return n;
                }
                else {
                    this.writerThread = NativeThread.current();
                    do {
                        send = this.send(this.fd, byteBuffer, checkAddress);
                    } while (send == -3 && this.isOpen());
                    synchronized (this.stateLock) {
                        if (this.isOpen() && this.localAddress == null) {
                            this.localAddress = Net.localAddress(this.fd);
                        }
                    }
                    int normalize = IOStatus.normalize(send);
                    this.writerThread = 0L;
                    this.end(send > 0 || send == -2);
                    assert !(!IOStatus.check(send));
                    return normalize;
                }
            }
            finally {
                this.writerThread = 0L;
                this.end(send > 0 || send == -2);
                assert !(!IOStatus.check(send));
            }
        }
    }
    
    private int send(FileDescriptor fileDescriptor, ByteBuffer src, InetSocketAddress inetSocketAddress) throws IOException {
        if ((src instanceof DirectBuffer)) {
            return this.sendFromNativeBuffer(fileDescriptor, src, inetSocketAddress);
        }
        int position = src.position();
        int limit = src.limit();
        assert position <= limit;
        ByteBuffer temporaryDirectBuffer = Util.getTemporaryDirectBuffer((position <= limit) ? (limit - position) : 0);
        try {
            temporaryDirectBuffer.put(src);
            temporaryDirectBuffer.flip();
            src.position(position);
            int sendFromNativeBuffer = this.sendFromNativeBuffer(fileDescriptor, temporaryDirectBuffer, inetSocketAddress);
            if (sendFromNativeBuffer > 0) {
                src.position(position + sendFromNativeBuffer);
            }
            return sendFromNativeBuffer;
        }
        finally {
            Util.releaseTemporaryDirectBuffer(temporaryDirectBuffer);
        }
    }
    
    private int sendFromNativeBuffer(FileDescriptor fileDescriptor, ByteBuffer byteBuffer, InetSocketAddress inetSocketAddress) throws IOException {
        int position = byteBuffer.position();
        int limit = byteBuffer.limit();
        assert position <= limit;
        int n = (position <= limit) ? (limit - position) : 0;
        int n2 = this.family != StandardProtocolFamily.INET;
        int send0;
        try {
            send0 = this.send0((boolean)n2, fileDescriptor, ((DirectBuffer)byteBuffer).address() + (long)position, n, inetSocketAddress.getAddress(), inetSocketAddress.getPort());
        }
        catch (PortUnreachableException ex) {
            if (this.isConnected()) {
                throw ex;
            }
            send0 = n;
        }
        if (send0 > 0) {
            byteBuffer.position(position + send0);
        }
        return send0;
    }
    
    @Override
    public int read(ByteBuffer byteBuffer) throws IOException {
        if (byteBuffer == null) {
            throw new NullPointerException();
        }
        synchronized (this.readLock) {
            synchronized (this.stateLock) {
                this.ensureOpen();
                if (!this.isConnected()) {
                    throw new NotYetConnectedException();
                }
            }
            int read = 0;
            try {
                this.begin();
                if (!this.isOpen()) {
                    int n = 0;
                    this.readerThread = 0L;
                    this.end(read > 0 || read == -2);
                    assert !(!IOStatus.check(read));
                    return n;
                }
                else {
                    this.readerThread = NativeThread.current();
                    do {
                        read = IOUtil.read(this.fd, byteBuffer, -1L, DatagramChannelImpl.nd);
                    } while (read == -3 && this.isOpen());
                    int normalize = IOStatus.normalize(read);
                    this.readerThread = 0L;
                    this.end(read > 0 || read == -2);
                    assert !(!IOStatus.check(read));
                    return normalize;
                }
            }
            finally {
                this.readerThread = 0L;
                this.end(read > 0 || read == -2);
                assert !(!IOStatus.check(read));
            }
        }
    }
    
    @Override
    public long read(ByteBuffer[] array, int n, int n2) throws IOException {
        if (n < 0 || n2 < 0 || n > array.length - n2) {
            throw new IndexOutOfBoundsException();
        }
        synchronized (this.readLock) {
            synchronized (this.stateLock) {
                this.ensureOpen();
                if (!this.isConnected()) {
                    throw new NotYetConnectedException();
                }
            }
            long read = 0L;
            try {
                this.begin();
                if (!this.isOpen()) {
                    long n3 = 0L;
                    this.readerThread = 0L;
                    this.end(read > 0L || read == -2L);
                    assert !(!IOStatus.check(read));
                    return n3;
                }
                else {
                    this.readerThread = NativeThread.current();
                    do {
                        read = IOUtil.read(this.fd, array, n, n2, DatagramChannelImpl.nd);
                    } while (read == -3L && this.isOpen());
                    long normalize = IOStatus.normalize(read);
                    this.readerThread = 0L;
                    this.end(read > 0L || read == -2L);
                    assert !(!IOStatus.check(read));
                    return normalize;
                }
            }
            finally {
                this.readerThread = 0L;
                this.end(read > 0L || read == -2L);
                assert !(!IOStatus.check(read));
            }
        }
    }
    
    @Override
    public int write(ByteBuffer byteBuffer) throws IOException {
        if (byteBuffer == null) {
            throw new NullPointerException();
        }
        synchronized (this.writeLock) {
            synchronized (this.stateLock) {
                this.ensureOpen();
                if (!this.isConnected()) {
                    throw new NotYetConnectedException();
                }
            }
            int write = 0;
            try {
                this.begin();
                if (!this.isOpen()) {
                    int n = 0;
                    this.writerThread = 0L;
                    this.end(write > 0 || write == -2);
                    assert !(!IOStatus.check(write));
                    return n;
                }
                else {
                    this.writerThread = NativeThread.current();
                    do {
                        write = IOUtil.write(this.fd, byteBuffer, -1L, DatagramChannelImpl.nd);
                    } while (write == -3 && this.isOpen());
                    int normalize = IOStatus.normalize(write);
                    this.writerThread = 0L;
                    this.end(write > 0 || write == -2);
                    assert !(!IOStatus.check(write));
                    return normalize;
                }
            }
            finally {
                this.writerThread = 0L;
                this.end(write > 0 || write == -2);
                assert !(!IOStatus.check(write));
            }
        }
    }
    
    @Override
    public long write(ByteBuffer[] array, int n, int n2) throws IOException {
        if (n < 0 || n2 < 0 || n > array.length - n2) {
            throw new IndexOutOfBoundsException();
        }
        synchronized (this.writeLock) {
            synchronized (this.stateLock) {
                this.ensureOpen();
                if (!this.isConnected()) {
                    throw new NotYetConnectedException();
                }
            }
            long write = 0L;
            try {
                this.begin();
                if (!this.isOpen()) {
                    long n3 = 0L;
                    this.writerThread = 0L;
                    this.end(write > 0L || write == -2L);
                    assert !(!IOStatus.check(write));
                    return n3;
                }
                else {
                    this.writerThread = NativeThread.current();
                    do {
                        write = IOUtil.write(this.fd, array, n, n2, DatagramChannelImpl.nd);
                    } while (write == -3L && this.isOpen());
                    long normalize = IOStatus.normalize(write);
                    this.writerThread = 0L;
                    this.end(write > 0L || write == -2L);
                    assert !(!IOStatus.check(write));
                    return normalize;
                }
            }
            finally {
                this.writerThread = 0L;
                this.end(write > 0L || write == -2L);
                assert !(!IOStatus.check(write));
            }
        }
    }
    
    @Override
    protected void implConfigureBlocking(boolean b) throws IOException {
        IOUtil.configureBlocking(this.fd, b);
    }
    
    public SocketAddress localAddress() {
        synchronized (this.stateLock) {
            return this.localAddress;
        }
    }
    
    public SocketAddress remoteAddress() {
        synchronized (this.stateLock) {
            return this.remoteAddress;
        }
    }
    
    @Override
    public DatagramChannel bind(SocketAddress socketAddress) throws IOException {
        synchronized (this.readLock) {
            synchronized (this.writeLock) {
                synchronized (this.stateLock) {
                    this.ensureOpen();
                    if (this.localAddress != null) {
                        throw new AlreadyBoundException();
                    }
                    InetSocketAddress checkAddress;
                    if (socketAddress == null) {
                        if (this.family == StandardProtocolFamily.INET) {
                            checkAddress = new InetSocketAddress(InetAddress.getByName("0.0.0.0"), 0);
                        }
                        else {
                            checkAddress = new InetSocketAddress(0);
                        }
                    }
                    else {
                        checkAddress = Net.checkAddress(socketAddress);
                        if (this.family == StandardProtocolFamily.INET && !(checkAddress.getAddress() instanceof Inet4Address)) {
                            throw new UnsupportedAddressTypeException();
                        }
                    }
                    SecurityManager securityManager = System.getSecurityManager();
                    if (securityManager != null) {
                        securityManager.checkListen(checkAddress.getPort());
                    }
                    Net.bind(this.family, this.fd, checkAddress.getAddress(), checkAddress.getPort());
                    this.localAddress = Net.localAddress(this.fd);
                }
            }
        }
        return this;
    }
    
    @Override
    public boolean isConnected() {
        synchronized (this.stateLock) {
            return this.state == 1;
        }
    }
    
    void ensureOpenAndUnconnected() throws IOException {
        synchronized (this.stateLock) {
            if (!this.isOpen()) {
                throw new ClosedChannelException();
            }
            if (this.state != 0) {
                throw new IllegalStateException("Connect already invoked");
            }
        }
    }
    
    @Override
    public DatagramChannel connect(SocketAddress socketAddress) throws IOException {
        synchronized (this.readLock) {
            synchronized (this.writeLock) {
                synchronized (this.stateLock) {
                    this.ensureOpenAndUnconnected();
                    InetSocketAddress checkAddress = Net.checkAddress(socketAddress);
                    SecurityManager securityManager = System.getSecurityManager();
                    if (securityManager != null) {
                        securityManager.checkConnect(checkAddress.getAddress().getHostAddress(), checkAddress.getPort());
                    }
                    if (Net.connect(this.family, this.fd, checkAddress.getAddress(), checkAddress.getPort()) <= 0) {
                        throw new Error();
                    }
                    this.state = 1;
                    this.remoteAddress = checkAddress;
                    this.sender = checkAddress;
                    this.cachedSenderInetAddress = checkAddress.getAddress();
                    this.cachedSenderPort = checkAddress.getPort();
                    this.localAddress = Net.localAddress(this.fd);
                    boolean blocking = false;
                    synchronized (this.blockingLock()) {
                        try {
                            blocking = this.isBlocking();
                            ByteBuffer allocate = ByteBuffer.allocate(1);
                            if (!(blocking == 0)) {
                                this.configureBlocking(false);
                            }
                            do {
                                allocate.clear();
                            } while (this.receive(allocate) != null);
                        }
                        finally {
                            if (!(blocking == 0)) {
                                this.configureBlocking(true);
                            }
                        }
                    }
                }
            }
        }
        return this;
    }
    
    @Override
    public DatagramChannel disconnect() throws IOException {
        synchronized (this.readLock) {
            synchronized (this.writeLock) {
                synchronized (this.stateLock) {
                    if (!this.isConnected() || !this.isOpen()) {
                        return this;
                    }
                    InetSocketAddress remoteAddress = this.remoteAddress;
                    SecurityManager securityManager = System.getSecurityManager();
                    if (securityManager != null) {
                        securityManager.checkConnect(remoteAddress.getAddress().getHostAddress(), remoteAddress.getPort());
                    }
                    disconnect0(this.fd, this.family == StandardProtocolFamily.INET6);
                    this.remoteAddress = null;
                    this.state = 0;
                    this.localAddress = Net.localAddress(this.fd);
                }
            }
        }
        return this;
    }
    
    private MembershipKey innerJoin(InetAddress maddr, NetworkInterface networkInterface, InetAddress inetAddress) throws IOException {
        if (!maddr.isMulticastAddress()) {
            throw new IllegalArgumentException("Group not a multicast address");
        }
        if ((maddr instanceof Inet4Address)) {
            if (this.family == StandardProtocolFamily.INET6 && !Net.canIPv6SocketJoinIPv4Group()) {
                throw new IllegalArgumentException("IPv6 socket cannot join IPv4 multicast group");
            }
        }
        else if ((maddr instanceof Inet6Address)) {
            if (this.family != StandardProtocolFamily.INET6) {
                throw new IllegalArgumentException("Only IPv6 sockets can join IPv6 multicast group");
            }
        }
        else {
            throw new IllegalArgumentException("Address type not supported");
        }
        if (inetAddress != null) {
            if (inetAddress.isAnyLocalAddress()) {
                throw new IllegalArgumentException("Source address is a wildcard address");
            }
            if (inetAddress.isMulticastAddress()) {
                throw new IllegalArgumentException("Source address is multicast address");
            }
            if (inetAddress.getClass() != maddr.getClass()) {
                throw new IllegalArgumentException("Source address is different type to group");
            }
        }
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager != null) {
            securityManager.checkMulticast(maddr);
        }
        synchronized (this.stateLock) {
            if (!this.isOpen()) {
                throw new ClosedChannelException();
            }
            if (this.registry == null) {
                this.registry = new MembershipRegistry();
            }
            else {
                MembershipKey checkMembership = this.registry.checkMembership(maddr, networkInterface, inetAddress);
                if (checkMembership != null) {
                    return checkMembership;
                }
            }
            MembershipKeyImpl.Type6 type6;
            if (this.family == StandardProtocolFamily.INET6 && (maddr instanceof Inet6Address || Net.canJoin6WithIPv4Group())) {
                int index = networkInterface.getIndex();
                if (index == -1) {
                    throw new IOException("Network interface cannot be identified");
                }
                byte[] inet6AsByteArray = Net.inet6AsByteArray(maddr);
                byte[] array = (inetAddress == null) ? null : Net.inet6AsByteArray(inetAddress);
                if (Net.join6(this.fd, inet6AsByteArray, index, array) == -2) {
                    throw new UnsupportedOperationException();
                }
                type6 = new MembershipKeyImpl.Type6(this, maddr, networkInterface, inetAddress, inet6AsByteArray, index, array);
            }
            else {
                Inet4Address anyInet4Address = Net.anyInet4Address(networkInterface);
                if (anyInet4Address == null) {
                    throw new IOException("Network interface not configured for IPv4");
                }
                int inet4AsInt = Net.inet4AsInt(maddr);
                int inet4AsInt2 = Net.inet4AsInt(anyInet4Address);
                int n = (inetAddress == null) ? 0 : Net.inet4AsInt(inetAddress);
                if (Net.join4(this.fd, inet4AsInt, inet4AsInt2, n) == -2) {
                    throw new UnsupportedOperationException();
                }
                type6 = new MembershipKeyImpl.Type4(this, maddr, networkInterface, inetAddress, inet4AsInt, inet4AsInt2, n);
            }
            this.registry.add(type6);
            return type6;
        }
    }
    
    @Override
    public MembershipKey join(InetAddress inetAddress, NetworkInterface networkInterface) throws IOException {
        return this.innerJoin(inetAddress, networkInterface, null);
    }
    
    @Override
    public MembershipKey join(InetAddress inetAddress, NetworkInterface networkInterface, InetAddress inetAddress2) throws IOException {
        if (inetAddress2 == null) {
            throw new NullPointerException("source address is null");
        }
        return this.innerJoin(inetAddress, networkInterface, inetAddress2);
    }
    
    void drop(MembershipKeyImpl membershipKeyImpl) {
        assert membershipKeyImpl.channel() == this;
        synchronized (this.stateLock) {
            if (!membershipKeyImpl.isValid()) {
                return;
            }
            try {
                if ((membershipKeyImpl instanceof MembershipKeyImpl.Type6)) {
                    MembershipKeyImpl.Type6 type5 = (MembershipKeyImpl.Type6)membershipKeyImpl;
                    Net.drop6(this.fd, type5.groupAddress(), type5.index(), type5.source());
                }
                else {
                    MembershipKeyImpl.Type4 type4 = (MembershipKeyImpl.Type4)membershipKeyImpl;
                    Net.drop4(this.fd, type4.groupAddress(), type4.interfaceAddress(), type4.source());
                }
            }
            catch (IOException detailMessage) {
                throw new AssertionError(detailMessage);
            }
            membershipKeyImpl.invalidate();
            this.registry.remove(membershipKeyImpl);
        }
        return;
    }
    
    void block(MembershipKeyImpl membershipKeyImpl, InetAddress inetAddress) throws IOException {
        assert membershipKeyImpl.channel() == this;
        assert membershipKeyImpl.sourceAddress() == null;
        synchronized (this.stateLock) {
            if (!membershipKeyImpl.isValid()) {
                throw new IllegalStateException("key is no longer valid");
            }
            if (inetAddress.isAnyLocalAddress()) {
                throw new IllegalArgumentException("Source address is a wildcard address");
            }
            if (inetAddress.isMulticastAddress()) {
                throw new IllegalArgumentException("Source address is multicast address");
            }
            if (inetAddress.getClass() != membershipKeyImpl.group().getClass()) {
                throw new IllegalArgumentException("Source address is different type to group");
            }
            int n;
            if ((membershipKeyImpl instanceof MembershipKeyImpl.Type6)) {
                MembershipKeyImpl.Type6 type5 = (MembershipKeyImpl.Type6)membershipKeyImpl;
                n = Net.block6(this.fd, type5.groupAddress(), type5.index(), Net.inet6AsByteArray(inetAddress));
            }
            else {
                MembershipKeyImpl.Type4 type4 = (MembershipKeyImpl.Type4)membershipKeyImpl;
                n = Net.block4(this.fd, type4.groupAddress(), type4.interfaceAddress(), Net.inet4AsInt(inetAddress));
            }
            if (n == -2) {
                throw new UnsupportedOperationException();
            }
        }
        return;
    }
    
    void unblock(MembershipKeyImpl membershipKeyImpl, InetAddress inetAddress) {
        assert membershipKeyImpl.channel() == this;
        assert membershipKeyImpl.sourceAddress() == null;
        synchronized (this.stateLock) {
            if (!membershipKeyImpl.isValid()) {
                throw new IllegalStateException("key is no longer valid");
            }
            try {
                if ((membershipKeyImpl instanceof MembershipKeyImpl.Type6)) {
                    MembershipKeyImpl.Type6 type5 = (MembershipKeyImpl.Type6)membershipKeyImpl;
                    Net.unblock6(this.fd, type5.groupAddress(), type5.index(), Net.inet6AsByteArray(inetAddress));
                }
                else {
                    MembershipKeyImpl.Type4 type4 = (MembershipKeyImpl.Type4)membershipKeyImpl;
                    Net.unblock4(this.fd, type4.groupAddress(), type4.interfaceAddress(), Net.inet4AsInt(inetAddress));
                }
            }
            catch (IOException detailMessage) {
                throw new AssertionError(detailMessage);
            }
        }
        return;
    }
    
    @Override
    protected void implCloseSelectableChannel() throws IOException {
        synchronized (this.stateLock) {
            if (this.state != 2) {
                DatagramChannelImpl.nd.preClose(this.fd);
            }
            ResourceManager.afterUdpClose();
            if (this.registry != null) {
                this.registry.invalidateAll();
            }
            long readerThread;
            if ((readerThread = this.readerThread) != 0L) {
                NativeThread.signal(readerThread);
            }
            long writerThread;
            if ((writerThread = this.writerThread) != 0L) {
                NativeThread.signal(writerThread);
            }
            if (!this.isRegistered()) {
                this.kill();
            }
        }
    }
    
    @Override
    public void kill() throws IOException {
        synchronized (this.stateLock) {
            if (this.state == 2) {
                return;
            }
            if (this.state == -1) {
                this.state = 2;
                return;
            }
            assert !this.isOpen() && !this.isRegistered();
            DatagramChannelImpl.nd.close(this.fd);
            this.state = 2;
        }
    }
    
    @Override
    protected void finalize() throws IOException {
        if (this.fd != null) {
            this.close();
        }
    }
    
    public boolean translateReadyOps(int n, int n2, SelectionKeyImpl selectionKeyImpl) {
        int nioInterestOps = selectionKeyImpl.nioInterestOps();
        int nioReadyOps = selectionKeyImpl.nioReadyOps();
        int n3 = n2;
        if ((n & Net.POLLNVAL) != 0x0) {
            return false;
        }
        if ((n & (Net.POLLERR | Net.POLLHUP)) != 0x0) {
            selectionKeyImpl.nioReadyOps(nioInterestOps);
            return (nioInterestOps & ~nioReadyOps) != 0x0;
        }
        if ((n & Net.POLLIN) != 0x0 && (nioInterestOps & 0x1) != 0x0) {
            n3 |= 0x1;
        }
        if ((n & Net.POLLOUT) != 0x0 && (nioInterestOps & 0x4) != 0x0) {
            n3 |= 0x4;
        }
        selectionKeyImpl.nioReadyOps(n3);
        return (n3 & ~nioReadyOps) != 0x0;
    }
    
    @Override
    public boolean translateAndUpdateReadyOps(int n, SelectionKeyImpl selectionKeyImpl) {
        return this.translateReadyOps(n, selectionKeyImpl.nioReadyOps(), selectionKeyImpl);
    }
    
    @Override
    public boolean translateAndSetReadyOps(int n, SelectionKeyImpl selectionKeyImpl) {
        return this.translateReadyOps(n, 0, selectionKeyImpl);
    }
    
    int poll(int n, long n2) throws IOException {
        assert Thread.holdsLock(this.blockingLock()) && !this.isBlocking();
        synchronized (this.readLock) {
            int poll = 0;
            try {
                this.begin();
                synchronized (this.stateLock) {
                    if (!this.isOpen()) {
                        int n3 = 0;
                        this.readerThread = 0L;
                        this.end(poll > 0);
                        return n3;
                    }
                    this.readerThread = NativeThread.current();
                }
                poll = Net.poll(this.fd, n, n2);
            }
            finally {
                this.readerThread = 0L;
                this.end(poll > 0);
            }
            return poll;
        }
    }
    
    @Override
    public void translateAndSetInterestOps(int n, SelectionKeyImpl selectionKeyImpl) {
        int n2 = 0;
        if ((n & 0x1) != 0x0) {
            n2 |= Net.POLLIN;
        }
        if ((n & 0x4) != 0x0) {
            n2 |= Net.POLLOUT;
        }
        if ((n & 0x8) != 0x0) {
            n2 |= Net.POLLIN;
        }
        selectionKeyImpl.selector.putEventOps(selectionKeyImpl, n2);
    }
    
    @Override
    public FileDescriptor getFD() {
        return this.fd;
    }
    
    @Override
    public int getFDVal() {
        return this.fdVal;
    }
    
    private static native void initIDs();
    
    private static native void disconnect0(FileDescriptor p0, boolean p1) throws IOException;
    
    private native int receive0(FileDescriptor p0, long p1, int p2, boolean p3) throws IOException;
    
    private native int send0(boolean p0, FileDescriptor p1, long p2, int p3, InetAddress p4, int p5) throws IOException;
    
    static {
        DatagramChannelImpl.nd = new DatagramDispatcher();
        IOUtil.load();
        initIDs();
    }
    
    private static class DefaultOptionsHolder
    {
        static final Set<SocketOption<?>> defaultOptions;
        
        private static Set<SocketOption<?>> defaultOptions() {
            HashSet s = new HashSet(8);
            s.add(StandardSocketOptions.SO_SNDBUF);
            s.add(StandardSocketOptions.SO_RCVBUF);
            s.add(StandardSocketOptions.SO_REUSEADDR);
            s.add(StandardSocketOptions.SO_BROADCAST);
            s.add(StandardSocketOptions.IP_TOS);
            s.add(StandardSocketOptions.IP_MULTICAST_IF);
            s.add(StandardSocketOptions.IP_MULTICAST_TTL);
            s.add(StandardSocketOptions.IP_MULTICAST_LOOP);
            s.addAll(ExtendedSocketOptions.datagramSocketOptions());
            return Collections.unmodifiableSet(s);
        }
        
        static {
            defaultOptions = defaultOptions();
        }
    }
}
